// $Id: SVGGraph.cc 3769 2010-10-28 22:38:42Z flaterco $

/*  SVGGraph  Graph implemented as Scalable Vector Graphics (SVG).

    Copyright (C) 2010 Jack Greenbaum and David Flater.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "xtide.hh"
#include "Graph.hh"
#include "SVGGraph.hh"


// FIXME:  boned on font metrics.
// - The default font is chosen by the user agent.
// - We must specify a font in order to know what the metrics are.
// - There is no font that is guaranteed to be available.
// - In Firefox, the actual size varies with scaling anyway, and is only
//   approximately right.

static const char     SVGfontName[]          = "DejaVu Sans";
static const unsigned SVGfontNominalSize     = 13U;
static const double   SVGfontBaselineAdjust  = 10.0;

// The following character widths are those claimed by Java's FontMetrics
// class when ostensibly the same font is loaded (under XP):
//   new Font ("DejaVu Sans", Font.PLAIN, 13)
// This has no reliable correlation with what we get as output from the SVG,
// but what's the alternative?

static const uint8_t SVGfontCharWidths[256] = {10,10,10,10,10,10,10,10,10,0,0,10,10,0,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,4,3,5,7,7,12,9,2,4,4,5,8,4,4,4,4,7,7,7,7,7,7,7,7,7,7,4,4,8,8,8,7,13,9,9,9,9,9,8,10,9,3,6,9,7,11,9,10,9,10,9,9,7,9,9,13,7,9,7,4,4,4,5,7,4,7,7,7,7,7,3,7,7,3,3,7,3,11,7,7,7,7,4,7,4,7,5,9,7,7,7,4,3,4,8,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,4,3,7,7,7,7,3,7,4,10,4,7,8,4,10,7,5,7,4,4,4,7,7,4,4,4,5,7,11,11,11,8,9,9,9,9,9,9,13,9,9,9,9,9,3,3,3,3,9,9,10,10,10,10,10,8,10,9,9,9,9,9,9,9,7,7,7,7,7,7,12,7,7,7,7,7,3,3,3,3,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};


SVGGraph::SVGGraph (unsigned xSize,
		    unsigned ySize,
		    GraphStyle style):
  Graph (xSize, ySize, style) {
  assert (xSize >= Global::minGraphWidth && ySize >= Global::minGraphHeight);

  SVG += "<?xml version=\"1.0\" standalone=\"no\"?>\n";
  SVG += "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n";

  SVG += "<svg width=\"";
  SVG += _xSize;
  SVG += "px\" height=\"";
  SVG += _ySize;
  SVG += "px\" viewBox=\"0 0 ";
  SVG += _xSize;
  SVG += ' ';
  SVG += _ySize;
  SVG += "\" xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\" xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n";

  // Translate color palette into SVG styles.
  SVG += "<defs>\n\t<style type=\"text/css\">\n\t\t<![CDATA[\n";
  for (unsigned a=0; a<Colors::numColors; ++a) {
    uint8_t r, g, b;
    Colors::parseColor (Global::settings[Colors::colorarg[a]].s, r, g, b);
    char SVGcolorSpec[8];
    snprintf (SVGcolorSpec, 8, "#%02x%02x%02x", r, g, b);
    SVG += "\t\t\t.";
    SVG += Colors::colorarg[a];
    SVG += " {";
#ifdef JGFeatures
    if (a == Colors::flood || a == Colors::ebb) {
      SVG += "opacity:";
      SVG += Global::settings["to"].d;
      SVG += ";";
    }
#endif
    SVG += "fill:";
    SVG += SVGcolorSpec;
    SVG += ";stroke:";
    SVG += SVGcolorSpec;
    SVG += "}\n";
    SVG += "\t\t\t.";
    SVG += Colors::colorarg[a];
    SVG += "nf {";
#ifdef JGFeatures
    if (a == Colors::flood || a == Colors::ebb) {
      SVG += "opacity:";
      SVG += Global::settings["to"].d;
      SVG += ";";
    }
#endif
    SVG += "fill:none;stroke:";
    SVG += SVGcolorSpec;
    SVG += "}\n";
  }
  SVG += "\t\t]]>\n\t</style>\n</defs>\n";

  // Establish clip box so labels at the edges don't run past the graphics.
  SVG += "<defs>\n\t<clipPath id=\"clipBox\">\n\t\t<rect x=\"0\" y=\"0\" width=\"";
  SVG += _xSize;
  SVG += "\" height=\"";
  SVG += _ySize;
  SVG += "\"/>\n\t</clipPath>\n</defs>\n<g clip-path=\"url(#clipBox)\" font-family=\"";
  SVG += SVGfontName;
  SVG += "\" font-size=\"";
  SVG += SVGfontNominalSize;
  SVG += "\">\n";
}


// FIXME:  stringWidth is bogus.
// Even if SVGfontCharWidths were accurate, the width of a string is not
// equal to the sum of the widths of its characters.  But what more can we
// do?
const unsigned SVGGraph::stringWidth (const Dstr &s) const {
  unsigned answer (0U);
  for (unsigned i=0; i<s.length(); ++i) {
    int normi = s[i];
    if (normi < 0)
      normi += 256;
    assert (normi >= 0 && normi < 256);
    answer += SVGfontCharWidths[normi];
  }
  // Add a wild guess at inter-character spacing, or a fudge factor
  answer += s.length() * 2;
  return answer;
}


const unsigned SVGGraph::fontHeight() const {
  return SVGfontNominalSize;
}


// Set to zero to slightly offset the enormous overestimate of stringWidth.
const unsigned SVGGraph::depthLabelLeftMargin() const {
  return 0;
}


void SVGGraph::drawX (double x, double y) {
  drawVerticalLineS   (x, y-4, y+4, Colors::foreground);
  drawHorizontalLineS (x-4, x+4, y, Colors::foreground);
}


void SVGGraph::drawBoxS (double x1, double x2, double y1, double y2,
			 Colors::Colorchoice c) {
  if (x1 > x2)
    std::swap (x1, x2);
  if (y1 > y2)
    std::swap (y1, y2);
  SVG += "<rect stroke-width=\"0\" x=\"";
  SVG += x1;
  SVG += "\" y=\"";
  SVG += y1;
  SVG += "\" width=\"";
  SVG += x2-x1;
  SVG += "\" height=\"";
  SVG += y2-y1;
  SVG += "\" class=\"";
  SVG += Colors::colorarg[c];
  SVG += "\"/>\n";
}


void SVGGraph::drawVerticalLineS (double x, double y1, double y2,
				  Colors::Colorchoice c) {
  SVG += "<line x1=\"";
  SVG += x;
  SVG += "\" y1=\"";
  SVG += y1;
  SVG += "\" x2=\"";
  SVG += x;
  SVG += "\" y2=\"";
  SVG += y2;
  SVG += "\" class=\"";
  SVG += Colors::colorarg[c];
  SVG += "\" stroke-width=\"1\"/>\n";
}


void SVGGraph::drawHorizontalLinePxSy (int xlo,
				       int xhi,
				       double y,
				       Colors::Colorchoice c) {
  if (xlo <= xhi) // Constraint per Graph::drawHorizontalLine.
    drawHorizontalLineS ((double)xlo, (double)xhi+1, y, c);
}


void SVGGraph::drawHorizontalLineS (double xlo, double xhi, double y,
				    Colors::Colorchoice c) {
  if (xlo <= xhi) { // Constraint per Graph::drawHorizontalLine.
    SVG += "<line x1=\"";
    SVG += xlo;
    SVG += "\" y1=\"";
    SVG += y;
    SVG += "\" x2=\"";
    SVG += xhi;
    SVG += "\" y2=\"";
    SVG += y;
    SVG += "\" class=\"";
    SVG += Colors::colorarg[c];
    SVG += "\" stroke-width=\"1\"/>\n";
  }
}


void SVGGraph::centerStringSxPy (double x, int y, const Dstr &s) {
  SVG += "<text x=\"";
  SVG += x;
  SVG += "\" y=\"";
  SVG += y + SVGfontBaselineAdjust;
  SVG += "\" text-anchor=\"middle\">";
  SVG += s;
  SVG += "</text>\n";
}


void SVGGraph::rightJustifyStringS (double x, double y, const Dstr &s) {
  SVG += "<text x=\"";
  SVG += x;
  SVG += "\" y=\"";
  SVG += y + SVGfontBaselineAdjust;
  SVG += "\" text-anchor=\"end\">";
  SVG += s;
  SVG += "</text>\n";
}


void SVGGraph::print (Dstr &text_out) {
  SVG += "</g>\n</svg>\n";
  text_out = SVG;
}


void SVGGraph::drawLevels (const SafeVector<double> &val,
			   const SafeVector<double> &y,
			   double yzulu,
			   bool isCurrent) {

  // FIXME:  drawLevels needs cleanup.

  // With JGFeatures, you get a semi-transparent filled curve topped with a
  // black line.  Otherwise, you get a filled curve or a multicolored line,
  // never both.

#ifndef JGFeatures
  if (Global::settings["nf"].c == 'n')
#endif

  // Filled curve drawing block
  {
    Colors::Colorchoice prevc=Colors::flood, c=Colors::flood;
    for (int x=0; x<=(int)_xSize; ++x) {
      prevc = c;
      if (isCurrent)
	c = (val[x+1] > 0.0 ? Colors::flood : Colors::ebb);
      else
	c = (val[x] < val[x+1] ? Colors::flood : Colors::ebb);
      if ((prevc != c) || (!x)) {
        // Start a new polygon, change colors.
	double xadj = x-1;
        // Don't cross yzulu with the wrong color.
	if (isCurrent && ((c == Colors::flood && y[x] > yzulu) ||
			  (c == Colors::ebb && y[x] < yzulu)))
	  xadj = linterp (xadj, (double)x, (yzulu-y[x])/(y[x+1]-y[x]));
        if (x) {
          SVG += ' ';
          SVG += xadj;
          SVG += ' ';
          if (isCurrent)
            SVG += yzulu;
          else
            SVG += _ySize;
          SVG += "\"/>\n";
        }
	SVG += "<polygon class=\"";
	SVG += Colors::colorarg[c];
	SVG += "\" stroke-width=\"0\" points=\"";
        if (isCurrent) {
          SVG += xadj;
          SVG += ' ';
          SVG += yzulu;
        } else {
          SVG += x-1;
          SVG += ' ';
          SVG += _ySize;
          SVG += ' ';
          SVG += x-1;
          SVG += ' ';
          SVG += y[x];
        }
      }
      SVG += ' ';
      SVG += x;
      SVG += ' ';
      SVG += y[x+1];
    }
    SVG += ' ';
    SVG += _xSize;
    SVG += ' ';
    if (isCurrent)
      SVG += yzulu;
    else
      SVG += _ySize;
    SVG += "\"/>\n";

  }

#ifndef JGFeatures
  else
#endif

  // Line-drawing block
  {
    Colors::Colorchoice prevc=Colors::foreground, c=Colors::foreground;
    const double lw (Global::settings["lw"].d);
    for (int x=0; x<=(int)_xSize+1; ++x) {
#ifndef JGFeatures
      prevc = c;
      if (isCurrent)
	c = (val[x+1] > 0.0 ? Colors::flood : Colors::ebb);
      else
	c = (val[x] < val[x+1] ? Colors::flood : Colors::ebb);
#endif
      if ((prevc != c) || (!x)) {
        // Start a new line, change colors.
        double xadj = x-1;
        double yadj = y[x];
        // Don't cross yzulu with the wrong color.
        if (isCurrent && ((c == Colors::flood && yadj > yzulu) ||
                          (c == Colors::ebb && yadj < yzulu))) {
          xadj = linterp (xadj, (double)x, (yzulu-yadj)/(y[x+1]-yadj));
          yadj = yzulu;
        }
        if (x) {
          if (isCurrent) {
	    SVG += ' ';
	    SVG += xadj;
	    SVG += ' ';
	    SVG += yzulu;
          }
          SVG += "\"/>\n";
        }
	SVG += "<polyline class=\"";
	SVG += Colors::colorarg[c];
	SVG += "nf\" stroke-width=\"";
	SVG += lw;
	SVG += "\" points=\"";
        SVG += xadj;
        SVG += ' ';
	SVG += yadj;
      }
      SVG += ' ';
      SVG += x;
      SVG += ' ';
      SVG += y[x+1];
    }
    SVG += "\"/>\n";
  }
}
